import logging
import re
import io
import copy
from scipy.stats import zscore
import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
import matplotlib
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.dates as mdates
import xlsxwriter
from bmxp.gravity import spearman, pearson
from bmxp import FMDATA, IMDATA
from .formation_plots import *

matplotlib.use("agg")

logging.basicConfig()
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)

__version__ = "0.3.4"


def parse_formatted(dataset):
    """
    Given a formatted dataset, parses it into the abundances, injection metadata,
    feature metadta, and the sample names
    """
    if isinstance(dataset, pd.DataFrame):
        df = dataset
    else:
        try:
            df = pd.read_csv(dataset, header=None)
        except Exception as e:  # pylint: disable=broad-except
            try:
                df = pd.read_excel(dataset, engine="openpyxl", header=None)
            except:
                raise ValueError(
                    "Your file does not appear to be a csv or xlsx."
                ) from e

    # Find the index of the first non-missing value
    first_row = df.iloc[0, :].copy()
    first_row.replace("", np.nan, inplace=True)
    pivot_col = df.columns.get_loc(first_row.first_valid_index())

    first_col = df.iloc[:, 0].copy()
    first_col.replace("", np.nan, inplace=True)
    pivot_row = df.index.get_loc(first_col.first_valid_index())

    if pivot_row == 0 and pivot_col == 0:
        # there's no imdata, use the last recognized fmdata header as the pivot
        for header in FMDATA.values():
            if header in first_row.values:
                pivot_col = max(pivot_col, first_row[first_row == header].index[-1])

    pivot_value = df.iloc[pivot_row, pivot_col].lower()

    if pivot_col < 0 or pivot_row < 0:
        raise ValueError("Your dataset is blank on either the first row or column.")

    feature_metadata = df.iloc[pivot_row + 1 :, : pivot_col + 1].copy()
    feature_metadata.columns = df.iloc[pivot_row, : pivot_col + 1]

    inj_metadata = df.iloc[: pivot_row + 1, pivot_col:].copy()
    inj_metadata = inj_metadata.transpose().reset_index(drop=True)
    inj_metadata.columns = inj_metadata.iloc[0]
    inj_metadata = inj_metadata.drop(0)
    inj_metadata.columns = inj_metadata.columns.str.lower()
    # makes column name unique - https://stackoverflow.com/questions/24685012
    cols = pd.Series(inj_metadata.columns)
    for dup in inj_metadata.columns[inj_metadata.columns.duplicated(keep=False)]:
        cols[inj_metadata.columns.get_loc(dup)] = [
            dup + "." + str(d_idx) if d_idx != 0 else dup
            for d_idx in range(inj_metadata.columns.get_loc(dup).sum())
        ]
    inj_metadata.columns = cols
    # check and create necessary columns

    sample_data = df.iloc[pivot_row + 1 :, pivot_col + 1 :]
    sample_names = df.iloc[pivot_row, pivot_col + 1 :].astype(str)
    sample_data.columns = inj_metadata.index.astype(str) + " (" + sample_names + ")"
    sample_data.index = feature_metadata.index.copy()
    sample_data = sample_data.replace("", "nan")
    sample_data = sample_data.astype(float)
    inj_metadata = inj_metadata.drop(pivot_value, axis=1)
    return sample_data, inj_metadata, feature_metadata, sample_names


def report(
    sample_data,
    smdata,
    fmdata,
    dataset_name,
    sample_names=None,
    out_filepath=None,
    write_pdf=True,
    warnings=None,
):
    """
    Handles /formation endpoint
    """
    fmdata = fmdata.copy()
    smdata = smdata.copy()
    sample_data = sample_data.copy()

    # raise ValueError("'Compound_ID' was not found in your dataset.")
    if "Compound_ID" in fmdata:
        fmdata.set_index("Compound_ID", drop=True, inplace=True)
        sample_data.index = fmdata.index

    if sample_names is None:
        if "reporting_name" in smdata:
            sample_names = smdata["reporting_name"].copy()
        elif "program_id" in smdata:
            sample_names = smdata["program_id"].copy()
        else:
            sample_names = pd.Series(smdata.index, index=smdata.index.copy())
    # check and create necessary columns

    fill_values = {
        "Annotation_ID": "",
        "HMDB_ID": "",
        "Metabolite": "",
    }
    for key, value in fill_values.items():
        if key not in fmdata.columns:
            fmdata[key] = value

    fmdata.columns.name = None

    if "raw_file_name" not in smdata.columns:
        smdata["raw_file_name"] = smdata.index
        # raise ValueError("'raw_file_name' was not found in the sample metadata.")

    fill_values = {
        "sample_type": "unknown_type",
        "injection_type": "unknown_type",
        "injection_order": list(range(len(smdata))),
        "column_number": 1,
    }
    for fill_key, value in fill_values.items():
        if fill_key not in smdata.columns:
            smdata[fill_key] = value

    # fill necessary missing values
    for sample_type in ("injection_type", "sample_type"):
        smdata[sample_type] = smdata[sample_type].fillna("")

    smdata.set_index("raw_file_name", drop=True, inplace=True)
    smdata = smdata.replace({"": float("nan"), "NA": float("nan")})
    smdata = smdata.convert_dtypes()
    # date_extracted usually does not convert to datetime correctly
    for date_type in ("date_extracted", "date_injected"):
        if date_type not in smdata.columns:
            continue
        try:
            smdata[date_type] = pd.to_datetime(smdata[date_type])
        except Exception as e:  # pylint: disable=broad-except
            raise ValueError(f"{date_type} contains one or more invalid dates.") from e

    try:
        smdata["injection_order"] = pd.to_numeric(smdata["injection_order"])
    except ValueError as e:
        character = re.search('".+"', str(e))
        if character:
            raise ValueError(
                f"Injection_order contains a non-numeric character: "
                f"{character.group()}."
            ) from e
        raise ValueError("Injection_order contains a non-numeric character.") from e

    transposed_s_data = sample_data.T
    transposed_s_data.index.name = "rawfile"
    transposed_s_data = transposed_s_data.apply(lambda x: x.fillna(x.min() / 2))
    # drop empty columns
    transposed_s_data = transposed_s_data.dropna(how="all", axis=1)
    s_data_zscores = zscore(transposed_s_data.to_numpy())
    s_data_zscores = np.nan_to_num(s_data_zscores)  # zscore is NaN if stdev is 0
    sample_pca = PCA(n_components=2)
    sample_pca_data = sample_pca.fit_transform(s_data_zscores)

    pca_df = pd.DataFrame(data=sample_pca_data, columns=["pc1", "pc2"])

    is_indices = fmdata.loc[fmdata["HMDB_ID"].str.lower() == "internal standard"].index
    internal_standard_s_data = sample_data.loc[is_indices]
    file_handle = io.BytesIO()
    palette = [
        # basel, from the R library 'yarrr'
        # https://cran.r-project.org/web/packages/yarrr/vignettes/piratepal.html
        (12, 91, 176, 0.7),  # blue
        (238, 0, 17, 0.7),  # red
        (21, 152, 61, 0.7),  # green
        (236, 87, 154, 0.7),  # pink
        (250, 107, 9, 0.7),  # orange
        (20, 155, 237, 0.7),  # light blue
        (161, 199, 32, 0.7),  # light green
        (254, 183, 11, 0.7),  # yellow
        (22, 160, 140, 0.7),  # teal
        (154, 112, 62, 0.7),  # brown
    ]
    nan_color = (0.55, 0.55, 0.55, 0.3)
    with PdfPages(file_handle) as pdf:
        plt.figure(figsize=(20, 12))
        plt.axis("off")
        plt.text(
            0.5,
            0.9,
            f"QC Report for {dataset_name}",
            ha="center",
            va="top",
            fontsize=24,
            wrap=True,
            in_layout=False,
        )
        plt.text(
            0.5,
            0.8,
            f"Number of samples: {len(sample_data.columns)}",
            ha="center",
            va="top",
            fontsize=18,
        )
        plt.text(
            0.5,
            0.7,
            f"Number of features: {len(fmdata.index)}",
            ha="center",
            va="top",
            fontsize=18,
        )
        if "date_injected" in smdata.columns:
            try:
                date = str(smdata["date_injected"].min().strftime("%B %d, %Y"))
            except:
                date = "Unknown"
            plt.text(
                0.5,
                0.6,
                ("Date of first injection: " + date),
                ha="center",
                va="top",
                fontsize=18,
            )
        if warnings:
            plt.text(
                0,
                0,
                "Warnings:\n"
                + "\n\n".join(
                    [
                        warning[:300] + "..." if len(warning) > 300 else warning
                        for warning in warnings
                    ]
                ),
                ha="left",
                va="bottom",
                fontsize=16,
                wrap=True,
                in_layout=False,
            )
        plt.tight_layout()
        pdf.savefig()
        plt.close("all")
        # create zoomable PCA plot
        plot_formation_zoomable_plot(
            pca_df,
            "PCA of metabolites labeled by sample name",
            f"PC1 ({round(sample_pca.explained_variance_ratio_[0] * 100)}"
            "% variance explained)",
            f"PC2 ({round(sample_pca.explained_variance_ratio_[1] * 100)}"
            "% variance explained)",
            sample_names,
            pdf,
        )

        components_df = pd.DataFrame(
            data=sample_pca.components_.transpose(),
            columns=["pc1", "pc2"],
            index=transposed_s_data.columns,
        )
        # create zoomable loadings plots
        plot_formation_zoomable_plot(
            components_df,
            "Loadings plot labeled by Annotation_ID",
            "Principal Component 1",
            "Principal Component 2",
            fmdata.loc[transposed_s_data.columns, "Metabolite"],
            pdf,
            rasterized=True,
        )
        plot_formation_zoomable_plot(
            components_df,
            "Loadings plot labeled by Compound_ID",
            "Principal Component 1",
            "Principal Component 2",
            transposed_s_data.columns,
            pdf,
            rasterized=True,
        )

        # create PCAs colored by inj metadata
        num_columns = smdata.select_dtypes(include=["float64", "int64"]).columns
        date_columns = smdata.select_dtypes(include=["datetime"]).columns
        str_columns = smdata.select_dtypes(include=["string"]).columns
        cmap = copy.copy(plt.cm.get_cmap("magma"))
        cmap.set_bad(color=nan_color)
        graph_index = 0
        for i, column_name in enumerate(smdata.columns):
            if i % 6 == 0 and i != 0:
                plt.tight_layout()
                pdf.savefig()
                plt.close("all")
            if i % 6 == 0:
                plt.figure(figsize=(20, 12))

            plt.subplot(2, 3, i % 6 + 1)
            plt.xlabel(
                f"PC1 ({round(sample_pca.explained_variance_ratio_[0] * 100)}"
                "% variance explained)"
            )
            plt.ylabel(
                f"PC2 ({round(sample_pca.explained_variance_ratio_[1] * 100)}"
                "% variance explained)"
            )
            plt.title("PCA of metabolites - colored by " + str(column_name))

            if column_name in num_columns:
                col = smdata[column_name].astype("float").replace({pd.NA: np.nan})
                plt.scatter(
                    pca_df["pc1"],
                    pca_df["pc2"],
                    c=col,
                    s=50,
                    cmap=cmap,
                    plotnonfinite=True,
                    alpha=0.7,
                )
                if col.isnull().values.any():
                    plt.legend(["NA"])
                    axes = plt.gca()
                    leg = axes.get_legend()
                    leg.legend_handles[0].set_color(nan_color)
                plt.colorbar(label=column_name)

            elif column_name in str_columns:
                col = smdata[column_name].replace({pd.NA: "NA"})
                targets = col.unique()
                colors = {
                    targets[k]: tuple(x / 255 for x in palette[k % len(palette)][:-1])
                    + (0.7,)
                    for k in range(len(targets))
                }
                plt.scatter(
                    pca_df["pc1"],
                    pca_df["pc2"],
                    color=col.map(colors),
                    s=50,
                )
                legend_colors = {k: colors[k] for k in list(colors)[:20]}
                handles = [
                    Line2D(
                        [0],
                        [0],
                        marker="o",
                        color="w",
                        markerfacecolor=v,
                        label=k,
                        markersize=8,
                    )
                    for k, v in legend_colors.items()
                ]
                leg = plt.legend(title="color", handles=handles)
                plt.gca().add_artist(leg)

            elif column_name in date_columns:
                plt.scatter(
                    pca_df["pc1"],
                    pca_df["pc2"],
                    c=mdates.date2num(smdata[column_name]),
                    s=50,
                    cmap=cmap,
                    alpha=0.7,
                )
                colorbar = plt.colorbar(label=column_name)
                loc = mdates.AutoDateLocator()
                colorbar.ax.yaxis.set_major_locator(loc)
                colorbar.ax.yaxis.set_major_formatter(mdates.ConciseDateFormatter(loc))
            graph_index = i
        # reset graph_index to create page break between PCAs and other plots
        graph_index = 0
        if not internal_standard_s_data.empty:
            is_ids = fmdata.loc[internal_standard_s_data.index, "Annotation_ID"]
            internal_standard_s_data.reset_index().apply(
                lambda row: plot_formation_line_plot(
                    row,
                    graph_index + 3 * row.name,
                    smdata,
                    palette,
                    pdf,
                    "injection_type",
                    ann_id=is_ids.iloc[row.name],
                ),
                axis=1,
            )
            graph_index += 3 * len(internal_standard_s_data.index)
            norm = internal_standard_s_data.apply(
                lambda row: row / row.median(), axis=1
            )
            line_colors = []
            num_inj_types = len(smdata["injection_type"].unique())
            for i in range(num_inj_types, num_inj_types + len(norm.index)):
                line_colors.append(
                    tuple(x / 255 for x in palette[i % len(palette)][:-1]) + (0.7,)
                )

            norm.reset_index().apply(
                lambda row: plot_formation_line_plot(
                    row,
                    graph_index,
                    smdata,
                    palette,
                    pdf,
                    "injection_type",
                    line_colors[row.name],
                    True,
                ),
                axis=1,
            )
            plt.ylim(bottom=0)
            leg_labels = [f"{k} - {v}" for k, v in zip(norm.index, is_ids)]
            handles_two = [
                Line2D([0], [0], color=v, label=k)
                for k, v in zip(leg_labels, line_colors)
            ]
            leg2 = plt.legend(title="color", handles=handles_two, loc=4)
            plt.gca().add_artist(leg2)
            graph_index += 3
        sample_medians = sample_data.fillna(0).median()
        plot_formation_line_plot(
            sample_medians,
            graph_index,
            smdata,
            palette,
            pdf,
            "sample_type",
            sample_names=sample_names,
        )
        graph_index += 3
        median_data = sample_data.apply(lambda col: col.fillna(col.min() / 2))
        median_data = median_data.apply(np.log)
        median_data = median_data.T
        unique_samp_types = list(smdata["sample_type"].unique())
        sample_colors = {
            key: tuple(x / 255 for x in value[:-1]) + (0.7,)
            for key, value in zip(unique_samp_types, palette)
        }
        colors = smdata["sample_type"].copy().astype(object).map(sample_colors)
        if len(median_data.index) <= 1500:
            for i in range(0, len(median_data.index), 150):
                plot_formation_quartile(
                    median_data.iloc[i : i + 150].T,
                    graph_index,
                    sample_colors,
                    colors[i : i + 150],
                    pdf,
                    sample_names[i : i + 150],
                )
                graph_index += 3
        plt.tight_layout()
        pdf.savefig()  # saves the current figure into a pdf page
        plt.close("all")
        inj_types = smdata["injection_type"].unique()
        pools = [
            inj_type.upper() for inj_type in inj_types if inj_type.startswith("pref")
        ]
        create_pools_cv_table(fmdata, pools, pdf)
        plt.tight_layout()
        pdf.savefig()  # saves the current figure into a pdf page
        plt.close("all")
    if out_filepath is None:
        out_filepath = f"{dataset_name}_formation_report.pdf"
    if write_pdf:
        with open(out_filepath, "wb") as f:
            f.write(file_handle.getbuffer())
    return file_handle


def report_from_formatted(
    dataset, dataset_name=None, out_filepath=None, write_pdf=True
):
    if dataset_name is None:
        if isinstance(dataset, pd.DataFrame):
            dataset_name = "Dataset"
        else:
            dataset_name = dataset.split("\\")[-1].split("/")[-1]

    sample_data, imdata, fmdata, sample_names = parse_formatted(dataset)

    return report(
        sample_data,
        imdata,
        fmdata,
        dataset_name,
        sample_names,
        out_filepath,
        write_pdf,
    )


def _sort_dataset(final_dataset):
    """
    Sort annotated metabolites by order, then class, MZ, RT; sort nontargeted by RT, MZ

    "orderNum" overrides all sorting rules except internal standard sorting; to skip
    internal standard sorting exclude "HMDB_ID" column or don't label internal standards
    """
    ordernum_present = "orderNum" in final_dataset
    if not ordernum_present:
        final_dataset["orderNum"] = np.nan
    # sort ordered or annotated features to top, then sort nontargeted by RT, MZ
    final_dataset = final_dataset.sort_values(by=["orderNum", "Metabolite", "RT", "MZ"])
    annotated = ~pd.isnull(final_dataset["Metabolite"]) | ~pd.isnull(
        final_dataset["orderNum"]
    )

    # force internal standards to top, missing orderNum next
    if "HMDB_ID" in final_dataset:
        final_dataset.loc[
            final_dataset["HMDB_ID"].str.lower() == "internal standard", "orderNum"
        ] = -99
    final_dataset.loc[annotated, "orderNum"] = final_dataset.loc[
        annotated, "orderNum"
    ].fillna(-98)

    # sort annotated if present
    try:
        final_dataset.loc[annotated, :] = (
            final_dataset.loc[annotated, :]
            .sort_values(
                by=[
                    "orderNum",
                    "superClass",
                    "mainClass",
                    "subClass",
                    "MZ_Calculated",
                    "RT",
                ],
                key=lambda col: (
                    col
                    if not pd.api.types.is_string_dtype(col.fillna(""))
                    else col.str.lower()
                ),
                na_position="first",
            )
            .values
        )
    except KeyError:
        pass

    if not ordernum_present:
        del final_dataset["orderNum"]

    return final_dataset


def harmonize_metadata(data, injectionset, sampleset):
    """
    Generates combined Sampleset and Injectionset metadata with columns matching data.
    If there are samples in data not present in either sampleset or injectionset, they
    are preserved and missing metadata is null.
    """
    warnings = []
    # rename column for older datasets
    if "reporting_name" not in injectionset.columns:
        injectionset = injectionset.rename(columns={"program_id": "reporting_name"})
    combined = injectionset.loc[data.columns].copy()
    combined["broad_id"] = combined["broad_id"].fillna("NA")

    s_broad_ids = combined.loc[combined["injection_type"] == "sample", "broad_id"]
    duplicates = set(s_broad_ids.loc[s_broad_ids.duplicated()]) - set([""])
    if len(duplicates) > 0:
        warnings.append(f"There are duplicate Broad_IDs: {', '.join(duplicates)}")

    sampleset_copy = sampleset.copy()
    sampleset_copy.loc["", :] = np.nan  # blank broad_id means no metadata
    combined = combined.join(sampleset_copy, on="broad_id", lsuffix="_inj")

    additional_meta = [
        col for col in sampleset.columns if col not in ["program_id", "broad_id"]
    ]

    # use injection set reporting_name for pools or missing program_ids if present
    combined["program_id"] = combined["program_id"].astype(object)
    pools = combined.index[combined["broad_id"].str.lower() == "pref"]
    missing = pd.isnull(combined["program_id"])
    if "reporting_name" in combined.columns:
        mismatches = combined.loc[
            (combined["program_id"] != combined["reporting_name"])
            & ~pd.isnull(combined["program_id"])
            & ~pd.isnull(combined["reporting_name"])
        ]
        if len(mismatches) > 0:
            warnings.append(
                "There are mismatches between program_ids and reporting_names. Please "
                + "review and confirm program_ids for the following samples before "
                + f"sharing results: {', '.join(mismatches.index)}"
            )
        combined.loc[missing, "program_id"] = combined.loc[missing, "reporting_name"]
    # name pools if possible
    elif pools.str.contains("PREF").all():
        combined.loc[pools, "program_id"] = pools.str.extract(r"(PREF\w*)").values
        warnings.append(
            "PREF program_id names were determined from injection_id names. Please "
            + "review and confirm program_id names before sharing results."
        )

    if pd.isnull(combined.loc[:, "program_id"]).any():
        warnings.append(
            "Not all program_ids could be filled. Please add any missing program_ids "
            "before sharing results."
        )
    combined[additional_meta] = combined[additional_meta].astype(object).fillna("NA")

    qcrole = IMDATA["QCRole"]
    qcrole_map = {
        "tube_blank": "QC-Tube_Blank",
        "blank": "QC-Blank",
        "not_used": "QC-Not_Used",
        "mm": "QC-Master_Mix",
        "ms2": "QC-MS2",
        "brpp": "QC-BRPP",
        "bridge_pref": "QC-Scaling_Pool",
    }

    if qcrole in combined.columns:
        for prefix, replacement in qcrole_map.items():
            mask = combined["injection_type"].str.startswith(prefix, na=False)
            combined.loc[mask, qcrole] = replacement
    else:
        combined[qcrole] = combined["injection_type"]

    # re-order
    injection_meta = combined.reindex(
        columns=[
            "date_extracted",
            "date_injected",
            "column_number",
            "injection_order",
            "injection_type",
            qcrole,
            "broad_id",
        ]
        + additional_meta
        + [
            "raw_file_name",
            "program_id",
        ]
    )
    injection_meta["raw_file_name"] = injection_meta.index
    # rename sample metadata and QCRole columns
    names = {col: col.lower().replace(" ", "_") for col in additional_meta}
    names[qcrole] = "sample_type"
    injection_meta = injection_meta.rename(columns=names)

    injection_meta.loc[:, "date_extracted"] = pd.to_datetime(
        injection_meta["date_extracted"],
        errors="coerce",
    )
    injection_meta.loc[:, "date_injected"] = pd.to_datetime(
        injection_meta["date_injected"],
        errors="coerce",
    )

    return injection_meta, warnings


def fill_f_mdata(metadata, annotations):
    """combine feature metadata from dataset and database queries for final results"""
    main_cols = ["__annotation_id", "Compound_ID", "MZ", "RT"]
    additional_meta = [col for col in metadata.columns if col not in main_cols]
    feature_meta = metadata.loc[:, main_cols + additional_meta]
    feature_meta["superClass"] = None
    feature_meta["mainClass"] = None
    feature_meta["subClass"] = None
    feature_meta["Adduct_Priority"] = None
    feature_meta["HMDB_ID"] = None
    feature_meta["HMDB_specificity (1=match; 2=representative)"] = None
    feature_meta["Annotation_ID"] = None
    feature_meta["Metabolite"] = None

    # move method to front
    method_col = ""
    if "Method" in feature_meta:
        method_col = feature_meta.pop("Method")
    feature_meta.insert(0, "Method", method_col)

    for key, value in annotations.items():
        if key not in feature_meta["__annotation_id"].unique():
            continue
        to_fill = feature_meta["__annotation_id"] == key
        for anno_key in value:
            feature_meta.loc[to_fill, anno_key] = value[anno_key]
    return feature_meta


def label_most_abundant(data, metadata, corr_method="spearman"):
    """Creates Primary and Corr_To_Primary columns given a clustered dataset. "Primary"
    feature is annotated or has the highest mean abundance.

    :param data: DataFrame, data columns that were used for clustering
    :param metadata: DataFrame, contains columns "Annotation_ID", "Cluster_Num", and
    "Cluster_Size" at minimum
    :param corr_method: str, correlation method used in clustering, "spearman" or
    "pearson", defaults to "spearman"
    :return: DataFrame, Primary and Corr_To_Primary columns
    """
    metadata = metadata.copy()
    # explicitly cast as str in case the Annotation_ID column is empty
    metadata["Annotation_ID"] = metadata["Annotation_ID"].fillna("")
    # identify the primary adduct
    for cluster_num in metadata["Cluster_Num"].unique():
        if not cluster_num and cluster_num != 0:
            continue
        cluster_index = metadata.index[metadata["Cluster_Num"] == cluster_num]
        key = (metadata["Cluster_Num"] == cluster_num).values & (
            metadata["Annotation_ID"] > ""
        ).values

        if key.any():
            metadata.loc[key, "Primary"] = True
            max_idx = metadata.index[key][0]
        else:
            max_idx = data.loc[cluster_index, :].fillna(0).mean(axis="columns").idxmax()
            metadata.loc[max_idx, "Primary"] = True

        for feature in cluster_index:
            arr1 = data.loc[feature, :].values.astype(np.float64)
            arr2 = data.loc[max_idx, :].values.astype(np.float64)
            if corr_method == "spearman":
                metadata.loc[feature, "Corr_To_Primary"] = spearman(
                    arr1, arr2, "drop", legacy_mode=True
                )
            else:
                metadata.loc[feature, "Corr_To_Primary"] = pearson(arr1, arr2)

    # label the singletons
    metadata.loc[metadata["Cluster_Size"] == 1, "Primary"] = True

    return metadata[["Primary", "Corr_To_Primary"]]


def apply_pool_instructions(data, feature_instructions, inplace=False):
    """Apply all pool instructions"""
    if feature_instructions is None:
        return data
    if not inplace:
        data = data.copy()
    for feature, instructions in feature_instructions.items():
        if feature not in data.index:
            continue
        for inst in instructions:
            if inst["op1"] not in data.columns or (
                inst["instruction"] != "delete" and inst["op2"] not in data.columns
            ):
                # we're probably doing final formatting and already removed some unused
                # pools, so just move on
                continue
            if inst["instruction"] == "swap":
                op1_value = data.loc[feature, inst["op1"]]
                data.loc[feature, inst["op1"]] = data.loc[feature, inst["op2"]]
                data.loc[feature, inst["op2"]] = op1_value
            elif inst["instruction"] == "copy":
                # copy op1 to op2
                data.loc[feature, inst["op2"]] = data.loc[feature, inst["op1"]]
            elif inst["instruction"] == "delete":
                data.loc[feature, inst["op1"]] = np.nan

    return data


def filter_samples_mask(data, smdata, formation_params):
    """
    Returns Boolean Mask of samples to keep
    the data columns must match the sample metadata index
    handles the old way, with 'only_prefs_samples' and 'keep_not_used'
    or the new way, with 'include_types'
    """
    warnings = []
    if not smdata.index.equals(data.columns):
        raise IndexError("Your Sample Metadata does not match your data columns.")

    if "include_types" in formation_params:
        if not formation_params["include_types"]:
            return np.array([True] * len(smdata)), warnings
        to_keep = np.array([False] * len(smdata))
        # Update to_keep for matching injection types
        for inj_type in formation_params["include_types"]:
            to_keep |= smdata["injection_type"].str.startswith(inj_type).values
    else:
        to_keep = np.array([True] * len(smdata))
        # filter out anything not labeled a pref or sample
        if formation_params["only_prefs_samples"]:
            to_keep = (
                smdata["injection_type"].str.startswith("pref").values
                | smdata["injection_type"].str.startswith("sample").values
            )

        # filter out not_used
        if not formation_params["keep_not_used"]:
            not_used = smdata["injection_type"].str.startswith("not_used").values
            to_keep = to_keep & ~not_used
    return to_keep, warnings


def filter_features_mask(data, imdata, fmdata, form_params):
    """
    Generates a filtering boolean key for data
    """
    warnings = []
    to_keep = np.array([True] * len(fmdata))
    ann_id_col = FMDATA["Annotation_ID"]
    adduct_col = FMDATA["Adduct"]
    inj_type_col = IMDATA["injection_type"]

    if not fmdata.index.equals(data.index):
        raise IndexError("Your Feature Metadata does not match your data index.")

    pref_as = imdata.index[imdata[inj_type_col].str.startswith("prefa", na=False)]
    pref_bs = imdata.index[imdata[inj_type_col].str.startswith("prefb", na=False)]

    if form_params.get("filter_by_pref_missing", False):
        miss_column = f"PREF Missing (of {len(pref_as) + len(pref_bs)})"
        if form_params.get("missing_as_percent", False):
            missing_cutoff = np.ceil(
                (len(pref_as) + len(pref_bs)) * form_params["missing_cutoff"] / 100
            )
        else:
            missing_cutoff = form_params["missing_cutoff"]
        to_keep = to_keep & (fmdata[miss_column] <= missing_cutoff)

    if form_params.get("filter_by_pref_cv", False):
        if len(pref_as) > 0:
            to_keep = to_keep & (fmdata["PREFA CVs"] <= form_params["cv_cutoff"] / 100)
        if len(pref_bs) > 0:
            to_keep = to_keep & (fmdata["PREFB CVs"] <= form_params["cv_cutoff"] / 100)

    if form_params.get("filter_by_clusters", False) and "Primary" in fmdata.columns:
        to_keep = to_keep & fmdata["Primary"].fillna(False)

    # annotated compounds don't get filtered by CV or missing PREFs
    to_keep = to_keep | pd.notnull(fmdata[ann_id_col])

    # but everything by non_quant
    if form_params.get("filter_by_nonquant", False):
        non_quant = fmdata[FMDATA["Non_Quant"]].fillna(False).astype(bool)
        to_keep = to_keep & ~non_quant

    # find all duplicated annotations that weren't filtered as non_quant
    annotated = to_keep & pd.notnull(fmdata[ann_id_col])
    is_duplicated = fmdata.loc[annotated, ann_id_col].duplicated(keep=False)
    duplicates = fmdata.loc[is_duplicated.index[is_duplicated]]
    keep_adducts = form_params.get("keep_adducts", "all")
    if adduct_col in fmdata.columns and keep_adducts != "all":
        if keep_adducts == "top":
            # only keep annotation if top priority adduct
            fmdata_adduct_priority_null_vals = (
                fmdata["Adduct_Priority"].fillna(False).astype(bool)
            )
            fmdata_adduct_priority_defined = fmdata[fmdata_adduct_priority_null_vals]
            fmdata_top_adduct = fmdata_adduct_priority_defined["Adduct_Priority"].map(
                lambda x: x.split(",")[0]
            )
            to_drop = fmdata_adduct_priority_defined[adduct_col] != fmdata_top_adduct
            to_drop_idx = to_drop.index[to_drop]
            to_keep[to_drop_idx] = False
        else:
            for anno in set(duplicates[ann_id_col]):
                group = duplicates.loc[duplicates[ann_id_col] == anno]
                # pick highest priority adduct, if possible
                priorities = group.loc[group.index[0], "Adduct_Priority"]
                if not priorities:
                    continue
                priorities = priorities.split(",")
                for adduct in priorities:
                    if adduct in group[adduct_col].values:
                        to_drop = group[adduct_col] != adduct
                        to_drop_idx = to_drop.index[to_drop]
                        to_keep[to_drop_idx] = False
                        break

    # find annotations that are still duplicated
    if form_params.get("filter_by_extraction_method", False):
        annotated = to_keep & pd.notnull(fmdata[ann_id_col])
        is_duplicated = fmdata.loc[annotated, ann_id_col].duplicated(keep=False)
        duplicates = fmdata.loc[is_duplicated.index[is_duplicated]]

        for anno in set(duplicates[ann_id_col]):
            group = duplicates.loc[duplicates[ann_id_col] == anno]
            # drop QI annotation(s) if there is a TF annotation
            if (
                form_params["feature_priority"][0]
                in group["__extraction_method"].values
            ):
                to_drop = (
                    group["__extraction_method"] == form_params["feature_priority"][1]
                )
                to_drop_idx = to_drop.index[to_drop]
                to_keep[to_drop_idx] = False

    # finally, if there are still duplicates, warn the user
    annotated = to_keep & pd.notnull(fmdata[ann_id_col])
    cols_to_check = [ann_id_col]
    if adduct_col in fmdata.columns and keep_adducts == "all":
        cols_to_check.append(adduct_col)
    is_duplicated = fmdata.loc[annotated].duplicated(subset=cols_to_check, keep=False)
    if sum(is_duplicated) > 0:
        warnings.append(
            "There are duplicate annotations that were not filtered by adduct "
            + "or extraction method. Please remove duplicate annotations before "
            + "sharing results."
        )

    return to_keep.values, warnings


def feature_qc(data, imdata, fmdata):
    """
    Generates QC data and adds to feature metadata
    """
    inj_type_col = IMDATA["injection_type"]
    pref_as = imdata.index[imdata[inj_type_col].str.startswith("prefa", na=False)]
    pref_bs = imdata.index[imdata[inj_type_col].str.startswith("prefb", na=False)]
    miss_column = f"PREF Missing (of {len(pref_as) + len(pref_bs)})"

    if len(pref_as) > 0:
        fmdata["PREFA CVs"] = data.loc[:, pref_as].std(axis=1) / data.loc[
            :, pref_as
        ].mean(axis=1)
    else:
        fmdata["PREFA CVs"] = None
    if len(pref_bs) > 0:
        fmdata["PREFB CVs"] = data.loc[:, pref_bs].std(axis=1) / data.loc[
            :, pref_bs
        ].mean(axis=1)
    else:
        fmdata["PREFB CVs"] = None
    fmdata[miss_column] = data.loc[:, pref_as.union(pref_bs)].isnull().sum(axis=1)
    return fmdata, miss_column


def prepare_sample_metadata(data, smdata):
    """
    Transposes sample metadata, aligns it with dataset columns, and creates column
    header DataFrame with program_ids
    """
    smdata = smdata.copy().T
    headers = data.columns.to_frame().T
    headers.loc[0, smdata.columns] = smdata.loc["program_id"]
    smdata = smdata.drop(index="program_id")
    smdata["Metabolite"] = smdata.index.str.capitalize()
    smdata = smdata.reindex(columns=data.columns, fill_value="")
    return smdata, headers


def combine_feature_metadata(data, fmdata, fmdata_formats, abundance_threshold=1):
    """
    Combines feature metadata and abundances into a single DataFrame, rounds data, and
    sorts features
    """
    warnings = []
    fmdata = fmdata.copy()

    # check fmdata and data have same index
    dataset_index = data.index
    if set(fmdata.index) != set(data.index):
        warnings.append(
            "Your metadata and data have mismatched features, likely because this"
            " is an old dataset. You might have missing features."
        )
        dataset_index = fmdata.index.intersection(data.index).copy()
        data = data.loc[dataset_index, :]
        fmdata = fmdata.loc[dataset_index, :]

    # sort and drop index (which was Compound_ID); Compound_ID should be in feature metadata
    fmdata = fmdata.loc[dataset_index, :].reset_index(drop=True)
    data.reset_index(drop=True, inplace=True)
    # fill non_quants and cast as object
    if "Non_Quant" not in fmdata.columns:
        fmdata["Non_Quant"] = False
    fmdata.fillna({"Non_Quant": False}, inplace=True)
    fmdata["Non_Quant"] = fmdata["Non_Quant"].astype("object")
    # remove values below abundance_threshold and round
    data[data < abundance_threshold] = np.nan
    data = data.round(0)
    data.replace(0, np.nan, inplace=True)
    # keep additional columns but drop columns that are for internal use only
    fmdata_cols = [col for col in fmdata_formats if col in fmdata.columns]
    additional_cols = [
        col for col in fmdata if col not in fmdata_cols and not col.startswith("__")
    ]
    fmdata = fmdata.reindex(columns=additional_cols + fmdata_cols)
    combined_dataset = pd.concat([fmdata, data], axis="columns", copy=False)
    combined_dataset = _sort_dataset(combined_dataset)
    combined_dataset = combined_dataset.reset_index(drop=True).reset_index()

    return combined_dataset, warnings


def to_excel(
    data,
    smdata,
    headers,
    fmdata_formats,
    smdata_formats,
    column_widths=None,
    method_name="Default",
    smformat_default="center",
    fmformat_default="header_center",
):
    """
    Returns a BytesIO object of a formatted excel sheet
    """

    output = io.BytesIO()
    workbook = xlsxwriter.Workbook(output)
    worksheet = workbook.add_worksheet(method_name)
    worksheet.add_write_handler(
        float,
        lambda worksheet, row, col, val, fmt=None: (
            None if not pd.isnull(val) else worksheet.write_blank(row, col, None, fmt)
        ),
    )
    formats = {
        "date": workbook.add_format(
            {
                "font_name": "Arial",
                "font_size": 9,
                "align": "center",
                "num_format": "m/d/yyyy",
            }
        ),
        "int": workbook.add_format(
            {
                "font_name": "Arial",
                "font_size": 9,
                "num_format": "0",
            }
        ),
        "center": workbook.add_format(
            {"font_name": "Arial", "font_size": 9, "align": "center"}
        ),
        "float2": workbook.add_format(
            {"font_name": "Arial", "font_size": 9, "num_format": "0.00"}
        ),
        "float4": workbook.add_format(
            {"font_name": "Arial", "font_size": 9, "num_format": "0.0000"}
        ),
        "label": workbook.add_format(
            {
                "font_name": "Arial",
                "font_size": 9,
                "italic": True,
                "align": "right",
            }
        ),
        "header": workbook.add_format(
            {"font_name": "Arial", "font_size": 9, "bold": True}
        ),
        "header_center": workbook.add_format(
            {"font_name": "Arial", "font_size": 9, "bold": True, "align": "center"}
        ),
        "default": workbook.add_format({"font_name": "Arial", "font_size": 9}),
        "percent": workbook.add_format(
            {"font_name": "Arial", "font_size": 9, "num_format": "0%"}
        ),
    }

    column_formats = {
        header: formats[label] for header, label in fmdata_formats.items()
    }
    row_formats = {index: formats[label] for index, label in smdata_formats.items()}
    smformat_default = formats[smformat_default]
    fmformat_default = formats[fmformat_default]

    # column widths
    worksheet.set_column_pixels(
        data.columns.get_loc("Compound_ID"),
        data.columns.get_loc("Compound_ID"),
        86,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("superClass"),
        data.columns.get_loc("superClass"),
        80,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("mainClass"),
        data.columns.get_loc("mainClass"),
        80,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("subClass"),
        data.columns.get_loc("subClass"),
        80,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("Annotation_ID"),
        data.columns.get_loc("Annotation_ID"),
        90,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("Non_Quant"),
        data.columns.get_loc("Non_Quant"),
        80,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("HMDB_ID"),
        data.columns.get_loc("HMDB_ID"),
        102,
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("HMDB_specificity (1=match; 2=representative)"),
        data.columns.get_loc("HMDB_specificity (1=match; 2=representative)"),
        270,
    )
    worksheet.set_column(
        data.columns.get_loc("Metabolite"),
        data.columns.get_loc("Metabolite"),
        data["Metabolite"].str.len().max(),
    )
    worksheet.set_column_pixels(
        data.columns.get_loc("Metabolite") + 1,
        len(data.columns) - 1,
        150,
    )
    if column_widths is not None:
        for col_name, width in column_widths.items():
            if col_name in data.columns:
                worksheet.set_column_pixels(
                    data.columns.get_loc(col_name),
                    data.columns.get_loc(col_name),
                    width,
                )

    for i, row in enumerate(smdata.values):
        index_label = smdata.index[i]
        if index_label in row_formats:
            worksheet.set_row_pixels(i, 16, row_formats[index_label])
        else:
            worksheet.set_row_pixels(i, 16, smformat_default)
        for j, val in enumerate(row):
            if data.columns[j] == "Metabolite":  # row labels
                worksheet.write(i, j, val, formats["label"])
            elif pd.isnull(val):
                # replace with a write handler for dates if we add more handlers later
                worksheet.write_blank(i, j, None)
            else:
                worksheet.write(i, j, val)

    row_offset = len(smdata)
    worksheet.set_row_pixels(row_offset, 16, formats["int"])
    for j, val in enumerate(headers.values[0]):
        if data.columns[j] in column_formats:  # metadata headers
            worksheet.write(row_offset, j, val, formats["header"])
        else:  # data headers
            worksheet.write(row_offset, j, val, fmformat_default)

    row_offset += 1
    for i, row in enumerate(data.values):
        index_label = data.index[i]
        worksheet.set_row_pixels(i + row_offset, 16, formats["int"])
        for j, val in enumerate(row):
            column_label = data.columns[j]
            if column_label in column_formats:  # feature metadata
                worksheet.write(i + row_offset, j, val, column_formats[column_label])
            else:
                worksheet.write(i + row_offset, j, val)

    workbook.close()
    return output



def interpolate_rts(combined, batch_num):
    # todo make sure this thing actually works
    rt_col = "RT"
    rt_new_col = f"RT_{batch_num}"
    rt_last_col = f"RT_Last_{batch_num}"

    combined = combined.copy()
    combined.sort_values(rt_col, inplace=True)
    combined.reset_index(inplace=True)  # We'll fix this later

    # Identify known and missing
    known = combined[combined[rt_new_col].notnull() & combined[rt_last_col].notnull()]
    missing = combined[combined[rt_new_col].isnull()]

    # For quick lookup
    known_rts = known[rt_col].to_numpy()
    known_diffs = (known[rt_new_col] - known[rt_last_col]).to_numpy()

    result_rt = combined[rt_new_col].to_numpy()
    result_last = combined[rt_last_col].to_numpy()

    for idx in missing.index:
        rt_val = combined.at[idx, rt_col]

        # Binary search range start and end (could optimize further)
        low = known_rts.searchsorted(rt_val - 0.2, side="right")
        high = known_rts.searchsorted(rt_val + 0.2, side="left")

        if low < high:
            nearby_diffs = known_diffs[low:high]
            avg_diff = np.nanmean(nearby_diffs)

            if not np.isnan(avg_diff):
                result_rt[idx] = rt_val
                result_last[idx] = rt_val - avg_diff

    # Write back updated values
    combined[rt_new_col] = result_rt
    combined[rt_last_col] = result_last

    still_null = pd.isnull(combined[rt_new_col])
    combined.loc[still_null, rt_new_col] = combined[rt_col]
    combined.loc[still_null, rt_last_col] = combined[rt_col]

    combined.set_index("index", inplace=True)  # restore original index

    return combined


# 5
def combine_batches(batches, start=0, stop=None, to_csv=True):
    """
    stop is not inclusive
    """
    MSAligner.descriptors = {"RT": "linear", "MZ": "ppm"}
    if stop is None:
        stop = len(batches)
    batch_nums = slice(start, stop)
    combined_sparse = None
    # combined_conserved = None
    prev_batch = None
    print("starting...", start, stop)
    lengths = {}
    qc_columns = [
        "Base_Gaussian_Correlation",
        "Prominence_Gaussian_Correlation",
        "Signal_To_Noise",
        "Best_Correlation",
        "FW20%M",
    ]

    important_columns = [
        "Compound_ID",
        "RT",
        "RT_Last",
        "MZ",
    ] + qc_columns
    intensity_column = "Intensity_Triangle"  # must be separate to avoid renaming
    lengths_sparse = {}
    combined_is_limited = False
    current_is_limited = False
    for batch_num in range(start, stop):
        print(batch_num)
        batch_columns = [col + f"_{batch_num}" for col in important_columns]
        try:
            current_batch = pd.read_csv(PB_DIR + DD_SLUG % batch_num)
        except:
            continue
        current_batch = current_batch[important_columns + [intensity_column]]

        if prev_batch is None:
            prev_batch = current_batch
            combined = current_batch.copy()
            combined["RT"] = combined["RT_Last"]
            continue
        print("calculating scalers...")
        # align to current to previous (*not* combined) to get scalers and stdevs
        a = MSAligner(
            prev_batch, current_batch, names=[str(batch_num - 1), str(batch_num)]
        )
        a.default_cutoff = 10
        a.default_coarse_params["RT"] = {
            "upper": 0.2,
            "lower": -0.2,
        }  # wider coarse matching
        a.default_scaler_params = {
            "smoothing_method": "lowess",
            "smoothing_params": {"frac": 0.5},
        }
        a.gen_anchors()
        a.gen_coarse()
        a.gen_scalers()
        a.gen_scaled_values()
        a.gen_stds()
        scalers = a.scalers
        stds = a.stds

        print("aligning to combined robust")
        # align current to robust combined features
        intensity_cols = [
            col for col in combined.columns if col.startswith("Intensity_Triangle")
        ]
        non_empty = combined[intensity_cols].notnull().all(axis=1)
        a = MSAligner(
            combined[non_empty],
            current_batch,
            names=[str(batch_num - 1), str(batch_num)],
        )

        a.default_cutoff = 10
        a.scalers = scalers
        a.stds = stds
        a.gen_scaled_values()
        a.gen_matches()
        a.gen_graph()
        results = a.results()  # ,c_size_or_loss=1, g_size_or_loss=1, diameter=1)
        print("parsing robust...")
        lengths[batch_num] = len(results)
        combined_robust = pd.DataFrame(index=results.index)

        # copy over all descriptors
        combined_robust.loc[:, important_columns + batch_columns] = results[
            important_columns + batch_columns
        ]
        # and all intensities
        intensity_cols = results.columns[
            results.columns.str.startswith("Intensity_Triangle")
        ]
        combined_robust.loc[:, intensity_cols] = results[intensity_cols]

        # get the sparse features which were aligned
        combined_robust_found = pd.Index(results[str(batch_num - 1)]).astype(int)

        # combined_robust_notfound = combined.index - combined_robust_found
        print("aligning to sparse")
        current_robust_found = pd.Index(results[str(batch_num)])
        current_robust_notfound = current_batch.index.difference(current_robust_found)
        a = MSAligner(
            combined[~non_empty],
            current_batch.loc[current_robust_notfound, :],
            names=[str(batch_num - 1), str(batch_num)],
        )

        a.default_cutoff = 10
        a.scalers = scalers
        a.stds = stds
        a.gen_scaled_values()
        a.gen_matches()
        a.gen_graph()
        results = a.results()
        print("parsing sparse")
        combined_sparse = pd.DataFrame(index=results.index)

        # copy over all descriptors for sparse
        combined_sparse.loc[:, important_columns + batch_columns] = results[
            important_columns + batch_columns
        ]
        intensity_cols = results.columns[
            results.columns.str.startswith("Intensity_Triangle")
        ]
        combined_sparse.loc[:, intensity_cols] = results[intensity_cols]

        # build the notfound databases
        combined_sparse_found = pd.Index(results[str(batch_num - 1)]).astype(int)
        current_sparse_found = pd.Index(results[str(batch_num)])

        combined_not_found = combined.index.difference(
            combined_robust_found
        ).difference(combined_sparse_found)
        print(combined.index[:5])
        print(combined_robust_found[:5])
        print(combined_sparse_found[:5])
        print(
            len(combined.index),
            len(combined_robust_found),
            len(combined_sparse_found),
            len(combined_not_found),
        )
        print(current_batch.index[:5])
        print(current_robust_found[:5])
        print(current_sparse_found[:5])

        current_not_found = current_batch.index.difference(
            current_robust_found
        ).difference(current_sparse_found)

        print(
            len(current_batch.index),
            len(current_robust_found),
            len(current_sparse_found),
            len(current_not_found),
        )
        df_combined_not_found = combined.loc[combined_not_found, :]
        df_current_not_found = current_batch.loc[current_not_found, :]
        print("interpolating1")
        # rename the current to add batch num so we can combine it with the others
        df_current_not_found = df_current_not_found.rename(
            columns={
                col: col + f"_{batch_num}"
                for col in important_columns + [intensity_column]
            }
        )

        combined = pd.concat(
            [
                combined_robust,
                combined_sparse,
                df_current_not_found,
                df_combined_not_found,
            ],
            ignore_index=True,
        )
        lengths_sparse[batch_num] = len(combined)

        # fill in missing MZ, RT_Last, Compound_ID for the most recent batch
        missing_mask = pd.isnull(combined[f"Intensity_Triangle_{batch_num}"])
        combined.loc[missing_mask, f"MZ_{batch_num}"] = combined.loc[missing_mask, "MZ"]
        combined.loc[missing_mask, f"Compound_ID_{batch_num}"] = combined.loc[
            missing_mask, "Compound_ID"
        ]
        print("interpolating2")
        combined = interpolate_rts(combined, batch_num)

        print("finalizing")
        combined["RT"] = combined["RT_Last"]

        # reassign columns to the "latest"
        for qc_col in qc_columns:
            combined[qc_col + f"_{batch_num}"] = combined[
                [qc_col, qc_col + f"_{batch_num}"]
            ].max(axis=1)
        intensity_cols = results.columns[
            results.columns.str.startswith("Intensity_Triangle")
        ]
        combined = combined[batch_columns + list(intensity_cols)]
        combined = combined.rename(
            columns={col: bcol for col, bcol in zip(batch_columns, important_columns)}
        )

        combined["RT"] = combined["RT_Last"]

        prev_batch = current_batch

    for k in lengths:
        print(f"{k}\t{lengths[k]}\t{lengths_sparse[k]}")

    # pull intensity and save
    only_results = pd.DataFrame(index=combined.index)
    only_results["Compound_ID"] = combined["Compound_ID"]
    only_results["RT"] = combined["RT"]
    only_results["MZ"] = combined["MZ"]
    only_results[qc_columns] = combined[qc_columns]
    intensity_columns = combined.columns[
        combined.columns.str.startswith("Intensity_Triangle")
    ]
    all_samples = list(
        dict.fromkeys(sample for batch in batches[batch_nums] for sample in batch)
    )
    only_results.loc[:, all_samples] = pd.NA

    for batch, int_col in zip(batches[batch_nums], intensity_columns):
        print("Batch: ", batch)
        json_parsed = []
        empty = [pd.NA] * len(batch)
        for i, (row_id, row_values) in enumerate(combined.iterrows()):
            if isinstance(row_values[int_col], str):
                json_parsed.append(json.loads(row_values[int_col]))
            else:
                json_parsed.append(empty)
        only_results.loc[:, batch] = json_parsed

    if to_csv:
        if not os.path.exists(COMB_DIR):
            os.makedirs(COMB_DIR)
        only_results.to_csv(COMB_DIR + "triangle.csv")
        combined.to_csv(COMB_DIR + COMB_SLUG)
    return only_results


def process(fwhm_min, fwhm_max):
    print("Processing with ", fwhm_min, fwhm_max)
    if not os.path.exists(RESULTS_DIR):
        os.makedirs(RESULTS_DIR)
    names, batches = get_batches(pd.read_csv(f"./injection_data.csv"))

    # get segments
    with multiprocessing.Pool(processes=os.cpu_count() - 2) as pool:
        pool.map(m_get_segments, names)
    #
    # identify single-file peaks
    partial_func = partial(m_segments_to_peaks, fwhm_min=fwhm_min, fwhm_max=fwhm_max)
    with multiprocessing.Pool(processes=os.cpu_count() - 2) as pool:
        pool.map(partial_func, names)

    # get combine peaks into a batch
    with multiprocessing.Pool(processes=10) as pool:
        args = [(batch, batch_id) for batch_id, batch in enumerate(batches)]
        pool.starmap(m_peaks_to_batches, args)

    # repick and integration batch peaks
    with multiprocessing.Pool(processes=10) as pool:
        args = [
            (batch, batch_id, fwhm_min, fwhm_max)
            for batch_id, batch in enumerate(batches)
        ]
        pool.starmap(m_pick_consensus, args)

    # combine batches
    return combine_batches(batches)



def interpolate_rts(combined, batch_num):
    # todo make sure this thing actually works
    """
    This fills in the missing RTs of the second batch with the first, based on consensus of matched in that region

    """
    ds1_rt_col = "RT"
    ds2_rt_col = f"RT_{batch_num}"
    combined = combined.copy()
    combined.sort_values(ds1_rt_col, inplace=True)
    combined.reset_index(inplace=True)  # We'll fix this later
    # Identify known and missing (from ds2)
    known = combined[combined[ds1_rt_col].notnull() & combined[ds2_rt_col].notnull()]
    missing = combined[combined[ds2_rt_col].isnull()]
    # For quick lookup
    known_rts = known[ds1_rt_col].to_numpy()
    known_diffs = (known[ds1_rt_col] - known[ds2_rt_col]).to_numpy()

    result_rt = combined[ds2_rt_col].to_numpy()

    # missing are those present in ds1 but not ds2
    for idx in missing.index:
        rt_val = combined.at[idx, ds1_rt_col]

        # Binary search range start and end (could optimize further)
        low = known_rts.searchsorted(rt_val - 0.2, side="right")
        high = known_rts.searchsorted(rt_val + 0.2, side="left")

        if low < high:
            nearby_diffs = known_diffs[low:high]
            avg_diff = np.nanmean(nearby_diffs)

            if not np.isnan(avg_diff):
                result_rt[idx] = rt_val - avg_diff

    # Write back updated values
    combined[ds2_rt_col] = result_rt
    still_null = pd.isnull(combined[ds2_rt_col])
    combined.loc[still_null, ds2_rt_col] = combined[ds1_rt_col]
    combined.set_index("index", inplace=True)  # restore original index
    return combined


# 5
def combine_batches(batches, start=0, stop=None, to_csv=True):
    """
    stop is not inclusive
    """
    MSAligner.descriptors = {"RT": "linear", "MZ": "ppm"}
    if stop is None:
        stop = len(batches)
    batch_nums = slice(start, stop)
    combined_sparse = None
    # combined_conserved = None
    prev_batch = None
    print("starting...", start, stop)
    lengths = {}
    qc_columns = [
        "Base_Gaussian_Correlation",
        "Prominence_Gaussian_Correlation",
        "Signal_To_Noise",
        "Best_Correlation",
        "FW20%M",
    ]

    important_columns = [
        "Compound_ID",
        "RT",
        "MZ",
    ] + qc_columns
    intensity_column = "Intensity_Triangle"  # must be separate to avoid renaming
    lengths_sparse = {}
    combined_is_limited = False
    current_is_limited = False
    for batch_num in range(start, stop):
        print(batch_num)
        batch_columns = [col + f"_{batch_num}" for col in important_columns]
        try:
            current_batch = pd.read_csv(PB_DIR + DD_SLUG % batch_num)
            current_batch["RT"] = current_batch["RT_Average"]
        except:
            continue

        current_batch = current_batch[important_columns + [intensity_column]]

        if prev_batch is None:
            prev_batch = current_batch
            combined = current_batch.copy()
            continue
        print("calculating scalers...")
        # align to current to previous (*not* combined) to get scalers and stdevs
        a = MSAligner(
            prev_batch, current_batch, names=[str(batch_num - 1), str(batch_num)]
        )
        a.default_cutoff = 10
        a.default_coarse_params["RT"] = {
            "upper": 0.2,
            "lower": -0.2,
        }  # wider coarse matching
        a.default_scaler_params = {
            "smoothing_method": "lowess",
            "smoothing_params": {"frac": 0.5},
        }
        a.gen_anchors()
        a.gen_coarse()
        a.gen_scalers()
        a.gen_scaled_values()
        a.gen_stds()
        scalers = a.scalers
        stds = a.stds
        a.report(filepath=COMB_DIR + f"batch_{batch_num}.pdf")

        print("aligning to combined robust")
        # align current to robust combined features
        intensity_cols = [
            col for col in combined.columns if col.startswith("Intensity_Triangle")
        ]
        non_empty = combined[intensity_cols].notnull().all(axis=1)
        a = MSAligner(
            combined[non_empty],
            current_batch,
            names=[str(batch_num - 1), str(batch_num)],
        )

        a.default_cutoff = 10
        a.scalers = scalers
        a.stds = stds
        a.gen_scaled_values()
        a.gen_matches()
        a.gen_graph()
        results = a.results()  # ,c_size_or_loss=1, g_size_or_loss=1, diameter=1)
        print("parsing robust...")
        lengths[batch_num] = len(results)
        combined_robust = pd.DataFrame(index=results.index)

        # copy over all descriptors
        combined_robust.loc[:, important_columns + batch_columns] = results[
            important_columns + batch_columns
        ]
        # and all intensities
        intensity_cols = results.columns[
            results.columns.str.startswith("Intensity_Triangle")
        ]
        combined_robust.loc[:, intensity_cols] = results[intensity_cols]

        # get the sparse features which were aligned
        combined_robust_found = pd.Index(results[str(batch_num - 1)]).astype(int)

        # combined_robust_notfound = combined.index - combined_robust_found
        print("aligning to sparse")
        current_robust_found = pd.Index(results[str(batch_num)])
        current_robust_notfound = current_batch.index.difference(current_robust_found)
        a = MSAligner(
            combined[~non_empty],
            current_batch.loc[current_robust_notfound, :],
            names=[str(batch_num - 1), str(batch_num)],
        )

        a.default_cutoff = 10
        a.scalers = scalers
        a.stds = stds
        a.gen_scaled_values()
        a.gen_matches()
        a.gen_graph()
        results = a.results()
        print("parsing sparse")
        combined_sparse = pd.DataFrame(index=results.index)

        # copy over all descriptors for sparse
        combined_sparse.loc[:, important_columns + batch_columns] = results[
            important_columns + batch_columns
        ]
        intensity_cols = results.columns[
            results.columns.str.startswith("Intensity_Triangle")
        ]
        combined_sparse.loc[:, intensity_cols] = results[intensity_cols]

        # build the notfound databases
        combined_sparse_found = pd.Index(results[str(batch_num - 1)]).astype(int)
        current_sparse_found = pd.Index(results[str(batch_num)])

        combined_not_found = combined.index.difference(
            combined_robust_found
        ).difference(combined_sparse_found)
        print(combined.index[:5])
        print(combined_robust_found[:5])
        print(combined_sparse_found[:5])
        print(
            len(combined.index),
            len(combined_robust_found),
            len(combined_sparse_found),
            len(combined_not_found),
        )
        print(current_batch.index[:5])
        print(current_robust_found[:5])
        print(current_sparse_found[:5])

        current_not_found = current_batch.index.difference(
            current_robust_found
        ).difference(current_sparse_found)

        print(
            len(current_batch.index),
            len(current_robust_found),
            len(current_sparse_found),
            len(current_not_found),
        )
        df_combined_not_found = combined.loc[combined_not_found, :]
        df_current_not_found = current_batch.loc[current_not_found, :]
        print("interpolating1")
        # rename the current to add batch num so we can combine it with the others
        df_current_not_found = df_current_not_found.rename(
            columns={
                col: col + f"_{batch_num}"
                for col in important_columns + [intensity_column]
            }
        )

        combined = pd.concat(
            [
                combined_robust,
                combined_sparse,
                df_current_not_found,
                df_combined_not_found,
            ],
            ignore_index=True,
        )
        lengths_sparse[batch_num] = len(combined)

        # fill in missing MZ, RT_Last, Compound_ID for the most recent batch
        # todo use m/z trends to interpolate mz
        missing_mask = pd.isnull(combined[f"Intensity_Triangle_{batch_num}"])
        combined.loc[missing_mask, f"MZ_{batch_num}"] = combined.loc[missing_mask, "MZ"]
        combined.loc[missing_mask, f"Compound_ID_{batch_num}"] = combined.loc[
            missing_mask, "Compound_ID"
        ]
        combined = interpolate_rts(combined, batch_num)

        print("finalizing")

        # reassign columns to the "latest"
        for qc_col in qc_columns:
            combined[qc_col + f"_{batch_num}"] = combined[
                [qc_col, qc_col + f"_{batch_num}"]
            ].max(axis=1)

        intensity_cols = results.columns[
            results.columns.str.startswith("Intensity_Triangle")
        ]
        # keep all the ds2 columns and rename
        combined = combined[batch_columns + list(intensity_cols)]
        combined = combined.rename(
            columns={col: bcol for col, bcol in zip(batch_columns, important_columns)}
        )

        prev_batch = current_batch

    for k in lengths:
        print(f"{k}\t{lengths[k]}\t{lengths_sparse[k]}")

    # pull intensity and save
    only_results = pd.DataFrame(index=combined.index)
    only_results["Compound_ID"] = combined["Compound_ID"]
    only_results["RT"] = combined["RT"]
    only_results["MZ"] = combined["MZ"]
    only_results[qc_columns] = combined[qc_columns]
    intensity_columns = combined.columns[
        combined.columns.str.startswith("Intensity_Triangle")
    ]
    all_samples = list(
        dict.fromkeys(sample for batch in batches[batch_nums] for sample in batch)
    )
    only_results.loc[:, all_samples] = pd.NA

    for batch, int_col in zip(batches[batch_nums], intensity_columns):
        print("Batch: ", batch)
        json_parsed = []
        empty = [pd.NA] * len(batch)
        for i, (row_id, row_values) in enumerate(combined.iterrows()):
            if isinstance(row_values[int_col], str):
                json_parsed.append(json.loads(row_values[int_col]))
            else:
                json_parsed.append(empty)
        only_results.loc[:, batch] = json_parsed

    if to_csv:
        if not os.path.exists(COMB_DIR):
            os.makedirs(COMB_DIR)
        only_results.to_csv(COMB_DIR + "triangle.csv")
        combined.to_csv(COMB_DIR + COMB_SLUG)
    return only_results



"""
A collection of python tools used for processing datasets at the
Metabolomics Platform at the Broad Institute. Includes
* Gravity - Clustering features based on RT and Correlation
* Eclipse - Alignment of nontargeted data
* Blueshift - Drift Correction based on internal standards and pooled references
* Formation - Formatting and QC of final datasets
"""

__version__ = "0.4.6"
# Feature Metadata
FMDATA = {
    # provided
    "Compound_ID": "Compound_ID",  # misnomer, project-unique feature ID
    "RT": "RT",  # unitless retention time, may or may not be scaled
    "MZ": "MZ",  # unsigned mass-to-charge ratio
    "Intensity": "Intensity",  # feature intensity
    "Annotation_ID": "Annotation_ID",  # method-unique annotation label
    "Adduct": "Adduct",  # the adduct form of the annotation
    "__annotation_id": "__annotation_id",  # globally unique annotation identifier
    "Metabolite": "Metabolite",  # preferred display/reporting name of metabolite
    "Non_Quant": "Non_Quant",  # boolean denoting that a feature is not quanitifiable
    # generated by gravity
    "Cluster_Num": "Cluster_Num",  # cluster number assigned during Gravity clustering
    "Cluster_Size": "Cluster_Size",  # number of members in the cluster
    # generated by blueshift
    "Batches Skipped": "Batches Skipped",  # batches that were skipped since there were no prefs
}

# injection or acquisition metadata
IMDATA = {
    "injection_id": "injection_id",  # injection name, usually filename
    "broad_id": "broad_id",  # internal biospecimen label
    "program_id": "program_id",  # external biospecimen label
    "injection_type": "injection_type",  # type of injection
    "notes": "notes",  # comments about the injection
    "column_number": "column_number",  # column number, relevant in multi-column studies
    "injection_order": "injection_order",  # injection number, not skipping blanks or non-samples
    "batches": "batches",  # 'batch_start' or 'batch_end'
    "reporting_name": "reporting_name",  # human readable reporting name for the injection
    # generated by blueshift
    "QCRole": "QCRole",  # "QC-drift_correction", "QC-pooled_ref", or "QC-not_used"
}

# sample metadata, not including
SMDATA = {
    "broad_id": "broad_id",  # internal biospecimen label
    "program_id": "program_id",  # external biospecimen label
}

# names of pooled reference injection types
POOL_INJ_TYPES = ("PREFA", "PREFB")



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import matplotlib.patches as mpatches


def plot_formation_zoomable_plot(
    pc_df, title, xaxis, yaxis, labels, pdf, rasterized=False
):
    """
    Creates large zoomable scatter plots for formation
    """
    fig = plt.figure(figsize=(20, 20), dpi=400)
    fig.set_rasterized(rasterized)
    plt.xlabel(xaxis)
    plt.ylabel(yaxis)
    plt.title(title)

    plt.scatter(pc_df["pc1"], pc_df["pc2"], c="#000", s=1, alpha=0.5)
    # add jitter to labels to prevent overlapping
    np.random.seed(0)
    pc1_stdev = 0.0005 * (max(pc_df["pc1"]) - min(pc_df["pc1"]))
    pc1_jitter = pc_df["pc1"] + np.random.randn(len(pc_df["pc1"])) * pc1_stdev
    pc2_stdev = 0.0005 * (max(pc_df["pc2"]) - min(pc_df["pc2"]))
    pc2_jitter = pc_df["pc2"] + np.random.randn(len(pc_df["pc2"])) * pc2_stdev

    for i, label in enumerate(labels):
        if pd.isnull(label):
            continue
        plt.annotate(
            label,
            xy=(pc_df["pc1"].iloc[i], pc_df["pc2"].iloc[i]),
            xycoords="data",
            xytext=(pc1_jitter.iloc[i], pc2_jitter.iloc[i]),
            textcoords="data",
            fontsize=2,
            ha="left",
            rotation=30,
        )
    plt.tight_layout()
    pdf.savefig()
    plt.close("all")


def plot_formation_line_plot(
    row,
    count,
    inj_metadata,
    palette,
    pdf,
    types,
    line_color="grey",
    is_aggregate=False,
    ann_id="",
    sample_names=None,
):
    """
    Creates line plot for formation
    """
    unique_types = list(inj_metadata[types].unique())
    colors = {}
    if not pd.isnull(unique_types[0]):
        if "sample" in unique_types:
            unique_types.remove("sample")
        if "Sample" in unique_types:
            unique_types.remove("Sample")
        colors = {
            key: tuple(x / 255 for x in value[:-1]) + (0.7,)
            for key, value in zip(unique_types, palette)
        }
        if "sample" in list(inj_metadata[types].unique()):
            colors["sample"] = (1, 1, 1, 0)
        if "Sample" in list(inj_metadata[types].unique()):
            colors["Sample"] = (1, 1, 1, 0)
    row_index = 2
    if (count % 6 == 0 or count % 6 > 3) and (not is_aggregate or row.name == 0):
        row_index = 1
        plt.tight_layout()
        pdf.savefig()
        plt.close("all")
        plt.figure(figsize=(20, 12))

    if not is_aggregate:
        plt.subplot(2, 1, row_index)
        if types == "injection_type":
            plt.title(
                ("Internal Standards Line Graph for: ")
                + (f"{row['Compound_ID']} - {ann_id}")
            )
            plt.ylabel("Raw Abundance")
        else:
            plt.title("Sample Abundance Median Plot")
            plt.ylabel("Median Abundance")

    elif row.name == 0:
        plt.subplot(2, 1, row_index)
        plt.title("Internal Standards Aggregate Line Graph")
        plt.ylabel("Normalized Abundance")

    plt.xlabel("Injection Order")

    if types == "injection_type":
        # remove internal standard name from values
        row_vals = row.values[1:]
    else:
        row_vals = row.values
    plt.scatter(
        inj_metadata["injection_order"],
        row_vals,
        c=inj_metadata[types].map(colors),
        zorder=2,
    )

    if "sample" in colors:
        del colors["sample"]
    if "Sample" in colors:
        del colors["Sample"]
    if not is_aggregate or (is_aggregate and row.name == 0):
        handles = [
            Line2D(
                [0],
                [0],
                marker="o",
                color="grey",
                markerfacecolor=v,
                label=k,
                markersize=8,
            )
            for k, v in colors.items()
        ]
        if types != "sample_type":
            leg_loc = 3 if is_aggregate else 0
            leg = plt.legend(title="color", handles=handles, loc=leg_loc)
            plt.gca().add_artist(leg)

    plt.plot(inj_metadata["injection_order"], row_vals, c=line_color, zorder=1)
    if not is_aggregate:
        plt.ylim(bottom=0)
    if types == "sample_type":
        handle_median_plot_labels(row, inj_metadata, sample_names, handles)


def handle_median_plot_labels(values, inj_metadata, sample_names, leg_handles):
    """Handles outliers and column break labels on Formation sample median plot"""
    combined_median_df = pd.concat(
        [
            inj_metadata.reset_index(drop=True)["injection_order"],
            values.reset_index(drop=True),
            sample_names.reset_index(drop=True),
        ],
        axis=1,
    )
    combined_median_df.columns = [
        "injection_order",
        "median_intensities",
        "sample_names",
    ]
    max_features = combined_median_df.nlargest(10, "median_intensities", keep="all")
    min_features = combined_median_df.nsmallest(10, "median_intensities", keep="all")
    min_max_features = pd.concat(
        [min_features, max_features], ignore_index=True, axis=0
    )
    np.random.seed(0)
    plt.scatter(
        min_max_features["injection_order"],
        min_max_features["median_intensities"],
        marker="*",
        zorder=3,
        color=(0, 0, 0, 0.5),
    )
    min_max_features.apply(
        lambda row: plt.annotate(
            row["sample_names"],
            xy=(row["injection_order"], row["median_intensities"]),
            xycoords="data",
            xytext=(np.random.randn() * 3, np.random.randn() * 5),
            textcoords="offset points",
            fontsize=5,
            ha="left",
            rotation=30,
        ),
        axis=1,
    )
    handles = leg_handles + [
        Line2D(
            [0],
            [0],
            marker="*",
            color="grey",
            markerfacecolor=(0, 0, 0, 0.5),
            label="10 Min and Max Samples",
            markersize=8,
        )
    ]
    leg = plt.legend(title="color", handles=handles)
    plt.gca().add_artist(leg)
    plt.ylim(bottom=0, top=1.05 * values.max())
    col_vals = inj_metadata.reset_index()["column_number"]
    col_change_vals = col_vals == col_vals.shift().fillna(col_vals)
    col_change = inj_metadata.reset_index()[~col_change_vals]
    col_change_x = col_change["injection_order"]
    plt.vlines(
        x=col_change_x,
        ymin=0,
        ymax=1.05 * values.max(),
        color="blue",
        linestyle="dashed",
    )


def plot_formation_quartile(data, count, sample_colors, colors, pdf, sample_names):
    """Creates formation sample boxplot"""
    row_index = 2
    if count % 6 == 0 or count % 6 > 3:
        row_index = 1
        plt.tight_layout()
        pdf.savefig()
        plt.close("all")
        plt.figure(figsize=(20, 12))
    axes = plt.subplot(2, 1, row_index)
    axes.set_rasterized(True)
    plt.title("Sample Feature Abundances (1/2 min imputed)")
    plt.xlabel("Sample")
    plt.ylabel("Log(Abundances)")

    medianprops = {"linestyle": "-.", "linewidth": 2.5, "color": "black"}
    bplot = plt.boxplot(data, widths=0.7, medianprops=medianprops, patch_artist=True)
    plt.xticks(range(1, len(data.columns) + 1), sample_names, rotation=90)
    for patch, color in zip(bplot["boxes"], colors):
        patch.set_facecolor(color)
    handles = [mpatches.Patch(color=v, label=k) for k, v in sample_colors.items()]
    axes.legend(
        title="sample_type", handles=handles, loc="center left", bbox_to_anchor=(1, 0.5)
    )


def create_pools_cv_table(feature_metadata, pools, pdf):
    """Creates formation Pools CV Table"""
    ann_features = feature_metadata.loc[~feature_metadata["Annotation_ID"].isnull()]
    bounds = [
        (0, 0.05),
        (0.05, 0.1),
        (0.1, 0.15),
        (0.15, 0.20),
        (0.20, 0.25),
        (0.25, 0.30),
        (0.30, 0.35),
        (0.35, 0.4),
        (0.4, ""),
    ]
    for i, pool in enumerate(pools):
        col_name = f"{pool} CVs"
        if col_name not in feature_metadata.columns:
            continue

        ann_features_col = ann_features[col_name].replace("", float("nan")).dropna()
        feature_met_col = feature_metadata[col_name].replace("", float("nan")).dropna()
        try:
            ann_features_col = ann_features_col.astype("float64")
            feature_met_col = feature_met_col.astype("float64")
        except ValueError:
            ann_indexer = ann_features_col.str.endswith("%")
            all_indexer = feature_met_col.str.endswith("%")
            ann_features_col.loc[ann_indexer] = ann_features_col.loc[
                ann_indexer
            ].str.strip("%")
            feature_met_col.loc[all_indexer] = feature_met_col.loc[
                all_indexer
            ].str.strip("%")
            ann_features_col = ann_features_col.astype("float64")
            feature_met_col = feature_met_col.astype("float64")
            ann_features_col.loc[ann_indexer] /= 100
            feature_met_col.loc[all_indexer] /= 100

        cv_vals = []
        rows_names = []
        for bound in bounds:
            if bound[1] == "":
                known_cvs = ann_features_col >= bound[0]
                all_cvs = feature_met_col >= bound[0]
                row_label = f">{round(bound[0] * 100)}%"
            else:
                known_cvs = (ann_features_col >= bound[0]) & (
                    ann_features_col < bound[1]
                )
                all_cvs = (feature_met_col >= bound[0]) & (feature_met_col < bound[1])
                row_label = f"{round(bound[0] * 100)}% - {round(bound[1] * 100)}%"
            known_count = known_cvs.sum()
            all_count = all_cvs.sum()

            cv_vals.append([known_count, all_count])
            rows_names.append(row_label)

        cv_vals.append(
            [
                ann_features[col_name].shape[0],
                feature_metadata[col_name].shape[0],
            ]
        )
        rows_names.append("Total")
        if i == 0:
            plt.figure(figsize=(20, 12))
        elif i % 4 == 0:
            plt.tight_layout()
            pdf.savefig()
            plt.close("all")
            plt.figure(figsize=(20, 12))
        plt.subplot(2, 2, i % 4 + 1)
        plt.gca().get_xaxis().set_visible(False)
        plt.gca().get_yaxis().set_visible(False)
        plt.box(on=None)
        plt.title(col_name)
        plt.table(
            cellText=cv_vals,
            rowLabels=rows_names,
            colLabels=["Knowns", "Knowns + Unknowns"],
            loc="center",
            cellLoc="center",
        )



# pylint: disable-all
import numpy as np
from bmxp.chroma import RawFile

if __name__ == "__main__":
    folder = "D:\\Work\\Tests\\RawFileReader\\"

    print("**** Testing Orbitrap MS1 ****")
    # Test full scan ms1 data
    alanine_rt = 7.73
    alanine_mz = 90.0554
    rt_dif = 0.02
    ppm = 5
    rts = np.array(
        [
            7.7116547,
            7.716111,
            7.720567,
            7.7250257,
            7.7294817,
            7.7339406,
            7.7383966,
            7.7428546,
            7.747311,
        ]
    )
    intensities = (
        np.array(
            [
                27994996.0,
                32120916.0,
                36518720.0,
                38417432.0,
                38131832.0,
                45260688.0,
                40830152.0,
                37549100.0,
                34188064.0,
            ]
        ),
    )

    pr_intensities = np.array(
        [
            9.6551744e07,
            1.1090329e08,
            1.2637120e08,
            1.3285288e08,
            1.3186667e08,
            1.5642437e08,
            1.4103850e08,
            1.2982436e08,
            1.1825154e08,
        ]
    )

    # test full scan ms1
    ms1_files = [
        "QE-ihmp2-rx0020a-hp.raw",
        "32-Centroid-zlib-NLC-QE.mzML",
        "32-Centroid-NoCompression-QE.mzML",
        "64-Centroid-NoCompression.mzML",
        "64-Centroid-NLC-QE.mzML",
        "64-Centroid-SLFC-QE.mzML",
        "64-Centroid-zlib-NLC-QE.mzML",
        "64-Centroid-zlib-QE.mzML",
    ]
    for ms1_file in ms1_files:
        print(ms1_file, " ", end="")
        profile = False
        centroid = True
        rf = RawFile(folder + ms1_file, profile=profile, centroid=centroid)
        xic = rf.xic(
            alanine_mz, alanine_rt - rt_dif, alanine_rt + rt_dif, ppm, centroid=centroid
        )
        print(
            np.isclose(xic.rt, rts).all(),
            np.isclose(
                xic.intensity,
                intensities,
            ).all(),
        )

    ms1_files = [
        "64-Profile-NoCompression.mzML",
        "32-Profile-NoCompression.mzML",
    ]
    for ms1_file in ms1_files:
        print(ms1_file, " ", end="")
        profile = True
        centroid = False
        rf = RawFile(folder + ms1_file, profile=profile, centroid=centroid)
        xic = rf.xic(
            alanine_mz, alanine_rt - rt_dif, alanine_rt + rt_dif, ppm, centroid=centroid
        )
        print(
            np.isclose(xic.rt, rts).all(),
            np.isclose(
                xic.intensity,
                pr_intensities,
            ).all(),
        )

    print("**** Testing QQQ Transitions****")
    # test triple quad
    filename = "Sciex-QTRAP-6500.mzML"
    print(filename, " ", end="")
    rf = RawFile(folder + filename)
    xic = rf.xic(369.4, 8.5, 8.6, precursor=675.2)
    print(np.isclose(xic.rt[0], 8.512), np.isclose(xic.intensity[0], 15604))

    # # test triple quad
    filename = "Agilent-6495-QQQ.mzML"
    print(filename, " ", end="")
    rf = RawFile(folder + filename)
    xic = rf.xic(44.296, 7.8, 7.9, precursor=90.06)
    print(np.isclose(xic.rt[0], 7.8011165), np.isclose(xic.intensity[0], 52846.855))



"""
Gravity clusters LCMS datasets by RT and Correlation


NaN Policy - Zeroes, Drop, Fill(spearman)

Method, NaN

Options:
Spearman - Fill NaNs with Zeroes (Fill in Python, call by_corr with ties=whatever)
Spearman - Drop NaNs (Don't fill, call by_rt with Ties=whatever)
Spearman - Backfill NaNs (Don't fill, call "backfill")

Pearson - Fill NaNs with Zeroes (Fill in Python, call by_corr)
Pearson - Drop NaNs (Don't fill, call by_rt)


"""

import logging
import math
from ctypes import CDLL, c_double, c_int32, c_void_p
import os
import platform
import pandas as pd
import numpy as np
import networkx as nx
from bmxp.gravity import fallback
from bmxp import FMDATA

dir_path = os.path.dirname(os.path.realpath(__file__))
correlation = None
if platform.system() == "Windows" and os.path.exists(
    os.path.join(dir_path, "correlation.dll")
):
    correlation = CDLL(os.path.join(dir_path, "correlation.dll"))
elif platform.system() == "Linux" and os.path.exists(
    os.path.join(dir_path, "correlation.so")
):
    correlation = CDLL(os.path.join(dir_path, "correlation.so"))

c_array = np.ctypeslib.ndpointer(dtype=np.float64, flags="C_CONTIGUOUS")
int32_array = np.ctypeslib.ndpointer(dtype=np.int32, flags="C_CONTIGUOUS")

if correlation:
    correlation.free_p.argtypes = [c_void_p]

    correlation.pearson_array.argtypes = [c_array, c_int32, c_int32, c_int32]
    correlation.pearson_array.restype = c_double

    correlation.spearman_array.argtypes = [
        c_array,  # arr
        c_int32,  # r1_start
        c_int32,  # r2_start
        c_int32,  # size
        c_int32,  # dropNan
        c_int32,  # legacyMode
    ]
    correlation.spearman_array.restype = c_double
    correlation.pearson.argtypes = [c_array, c_array, c_int32]
    correlation.pearson.restype = c_double

    correlation.spearman.argtypes = [c_array, c_array, c_int32, c_int32, c_int32]
    correlation.spearman.restype = c_double
else:
    correlation = fallback

logging.basicConfig()
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)

__version__ = "0.2.0"


def free_p(p):
    correlation.free_p(p)


def pearson_array(arr1, r1_start, r2_start):
    r_p = correlation.pearson_array(
        arr1,
        r1_start,
        r2_start,
        len(arr1[0]),
    )
    return r_p


def spearman_array(arr1, r1_start, r2_start, nan_policy="fill", legacy_mode=False):
    drop_nan = nan_policy != "fill"
    r_p = correlation.spearman_array(
        arr1, r1_start, r2_start, len(arr1[0]), drop_nan, legacy_mode
    )
    return r_p


def pearson(x, y):
    x = np.array(x).astype(np.float64)
    y = np.array(y).astype(np.float64)
    return correlation.pearson(x, y, len(x))


def spearman(x, y, nan_policy="fill", legacy_mode=False):
    x = np.array(x).astype(np.float64)
    y = np.array(y).astype(np.float64)
    drop_nan = nan_policy != "fill"
    return correlation.spearman(x, y, len(x), drop_nan, legacy_mode)


def deconvolute(df_index, graph, fmdata):
    num_label = fmdata["Cluster_Num"]
    size_label = fmdata["Cluster_Size"]
    cluster_df = pd.DataFrame({num_label: None, size_label: 1}, index=df_index)
    i = 0
    while True:
        # delete singletons
        graph.remove_nodes_from(list(nx.isolates(graph)))

        # build list of cliques of the largest size
        cliques = []
        max_size = 0
        for clique in nx.find_cliques(graph):
            if len(clique) > max_size:
                max_size = len(clique)
                cliques = [[set(clique), 0]]
            elif len(clique) == max_size:
                cliques.append([set(clique), 0])
        if not cliques:
            break
        # calculate correlation of largest cliques
        for clique in cliques:
            clique_graph = graph.subgraph(clique[0])
            temp = 0
            for edge in clique_graph.edges(data=True):
                temp += edge[2]["coor"]
            clique[1] = temp

        # sort by best corr
        cliques.sort(key=lambda x: x[1], reverse=True)
        to_remove = set()
        while len(cliques) > 0:
            # add the best one
            best_clique = cliques[0]
            features = list(best_clique[0])
            cluster_df.loc[features, num_label] = i
            cluster_df.loc[features, size_label] = max_size
            if i % 50 == 0:
                LOGGER.info(f"On cluster {i}, which contains {max_size} features.")
            i += 1

            # search the clique for ones to remove
            cliques = [
                clique
                for clique in cliques
                if not clique[0].intersection(best_clique[0])
            ]
            # add to our "to remove" queue
            to_remove.update(best_clique[0])
        graph.remove_nodes_from(to_remove)
        if max_size == 2:
            break
    return cluster_df


def corr_array(
    i, j, rt_series, df, corr_value, rt_thresh, method, nan_policy, legacy_mode
):
    # long winded way to find RT differences and the indices
    # similar to pd.stack
    np_batch_a = rt_series.iloc[i[0] : i[1]].values
    np_batch_b = rt_series.iloc[j[0] : j[1]].values
    rt_flattened = np.absolute(np.subtract.outer(np_batch_b, np_batch_a)).flatten()
    a_index = np.tile(np.fromiter(range(i[0], i[1]), int), (j[1] - j[0]))
    b_index = np.tile(np.fromiter(range(j[0], j[1]), int), ((i[1] - i[0]), 1)).flatten(
        order="F"
    )
    to_keep = (rt_flattened < rt_thresh) & (a_index != b_index)
    a_index = a_index[to_keep]
    b_index = b_index[to_keep]
    corr_results = np.empty(len(a_index))
    for k, a_val in enumerate(a_index):
        b_val = b_index[k]
        if method == "spearman":
            corr_results[k] = spearman_array(df, a_val, b_val, nan_policy, legacy_mode)
        else:
            corr_results[k] = pearson_array(df, a_val, b_val)
    to_return = corr_results > corr_value
    return a_index[to_return], b_index[to_return], corr_results[to_return]


def gen_batches(num_features, batch_size):
    """
    Generates batches for clustering, based on number of features and batch size
    """
    num_batches = math.ceil(num_features / batch_size)
    for i in range(num_batches):
        for j in range(i, num_batches):
            i_indices = [
                i * batch_size,
                min((i + 1) * batch_size, num_features),
            ]
            j_indices = [
                j * batch_size,
                min((j + 1) * batch_size, num_features),
            ]
            yield {
                "i": i,
                "j": j,
                "i_indices": i_indices,
                "j_indices": j_indices,
                "num_batches": num_batches,
            }


def cluster(
    df,
    rt_thresh=0.02,
    corr_value=0.8,
    batch_size=1000,
    method="spearman",
    nan_policy="fill",
    legacy_mode=False,
    schema_labels=None,
):
    """
    Cluster aggregates LCMS features into groups based on sample-correlation and
        retention time. It builds a network based on retention time difference and
        correlation. Then it labels the largest clique as a cluster, deletes the
        features from the network, and identifies the next largest clique. Ties are
        broken but a summation of correlations in the cluster.

    df: Dataframe, Eclipse compatible
    rt_thresh: float, maximum Rt different for clustering
    corr_value: float, correlation cutoff
    batch_size: int, batch size for calculating correlations
    method: string, "Spearman" or "Pearson" correlation
    legacy_mode: boolean, when True, NaNs are filled with 0s when correlating features
    with very few overlapping non-NaN values (Spearman only)
    Returns
    Dataframe, containing cluster number and number of members for each feature. -1
       indicates a single, unclustered feature.
    """
    fmdata = FMDATA.copy()
    if schema_labels is not None:
        fmdata.update(schema_labels)
    rt_series = df[fmdata["RT"]]
    rt_series.index.name = None  # otherwise it crashes during stack
    sample_df = df.copy().drop(list(fmdata.keys()), axis=1, errors="ignore")

    num_features = len(sample_df.index)
    graph = nx.Graph()
    sample_df = sample_df.values.copy()
    if nan_policy == "zeroes":
        sample_df = np.nan_to_num(sample_df)
    for batch_info in gen_batches(num_features, batch_size):
        num_batches = batch_info["num_batches"]
        LOGGER.info(f"Correlating Batch { batch_info['i'] + 1}/{num_batches}...")
        a_index, b_index, corrs = corr_array(
            batch_info["i_indices"],
            batch_info["j_indices"],
            rt_series,
            sample_df,
            corr_value,
            rt_thresh,
            method,
            nan_policy,
            legacy_mode,
        )
        a_index = df.index[a_index]
        b_index = df.index[b_index]
        edges = [(s, t, {"coor": coor}) for s, t, coor in zip(a_index, b_index, corrs)]
        graph.add_nodes_from(a_index)
        graph.add_nodes_from(b_index)
        graph.add_edges_from(edges)
    return deconvolute(df.index, graph, fmdata)


if __name__ == "__main__":
    data = pd.read_csv(
        r"C:\Users\danie\work\beta-muricholate.csv", index_col="Compound_ID"
    )
    data = data.replace(0, np.nan)
    # fmt: off
    arr1 = [
        [np.nan, np.nan],  # 0
        [1, np.nan],  # 1
        [1, np.nan],
        [1, np.nan],
        [1, np.nan],
        [2, np.nan],  # 5
        [2, np.nan],
        [2, 1],  # 7
        [np.nan, 2],
        [3, np.nan],
        [3, np.nan],  # 10
        [3, 5],
        [np.nan, np.nan],
        [4, 7],
        [4, np.nan],
        [4, 9],  # 15
        [4, 10],  # 16
        [np.nan, np.nan],  # 17
    ]

    arr = np.array(arr1).T.astype(np.float64).copy()
    # fmt: on

    # print(pearson_array_m(arr1, 0, 1))
    # spearman zeroes
    # spearman fill
    # spearman drop
    # pearson zeroes
    # pearson drop
    print(spearman_array(np.nan_to_num(arr), 0, 1, nan_policy="drop"))
    # print(pearson_array(arr1, 0, 1))
    # print("done")



"""Module from __init__.py"""

from bmxp.eclipse import MSAligner


datasets = [
    "../../tests/test1.csv",
    "../../tests/test2.csv",
    "../../tests/example1.csv",
]
names = ["DS1", "DS2", "DS3"]
############
# Example 1 - Unsupervised alignment between all datasets
############

a = MSAligner(*datasets, names=names)
a.align()
# Automatically aligns all datasets to all other datasets with defaults

print(a.anchors["DS1"])
# Output list of features
# ['0', '1', '10', '100', '101', ...]

print(a.coarse_matches["DS1"]["DS2"])
# Output the coarse alignment results from DS1 -> DS2.
# Will contain duplicates.
#      DS1  DS2
# 0      0    0
# 1      1    1
# 2     10   10
# 3    100  100
# 4    101  101


print(a.scalers["DS1"]["DS2"])
# Output dict of scalers. For example:
# [271 rows x 2 columns]
# {'RT':         x         y
# 0    1.68  0.004317
# 1    3.53 -0.008928
# ..    ...       ...
# 269  5.51 -0.021388
# 270  5.45 -0.020587
#
# [271 rows x 2 columns], 'MZ':             x         y
# 0    121.1013  0.828166
# 1    197.0667  1.028160
# ..        ...       ...
# 269  792.5513  1.144283
# 270  776.5563  1.152530
#
# [271 rows x 2 columns], 'Intensity':             x         y
# 0    4.703667 -0.097755
# 1    4.341406 -0.167542
# ..        ...       ...
# 269  5.963024 -0.089592
# 270  6.352471 -0.094823


print(a.stds["DS1"]["DS2"])
# Output standard deviations to DS2
# {
#   'RT': 0.01441794266731159,
#   'MZ': 0.12278117619894874,
#   'Intensity': 0.1113562600674442
# }


print(a.scaled_values["DS1"]["DS2"])
# Output DataFrame of scaled values
#            RT          MZ     Intensity
# 0    1.684317  121.101400  4.035640e+04
# 1    3.521072  197.066903  1.492328e+04
# ..        ...         ...           ...
# 297  7.319298  793.542507  1.170663e+06
# 298  0.345388  158.026900  1.156930e+06


print(a.matches["DS1"]["DS2"])
# Output a DataFrame of DS1 feature names to top n (1? 2?) DS2 matches
# [299 rows x 3 columns]
#         0
# 0       0
# 1    None
# 2       2
# ..    ...
# 296  None
# 297   268
# 298  None


# Saves a report of the alignment params
a.report(filepath="results.pdf")

# Reports only matches in all 3 datasets
a.to_csv(filepath="full_matches.csv")

# Reports all cliques with a match to dataset 1
# (Will likely have multiple matches)
a.to_csv(
    "DS1", filepath="DS1.csv", g_size_or_loss=1, c_size_or_loss=1, remove_rerank=False
)


############
# Example 2 - Supervised all by all alignment with custom initial params
############
a = MSAligner(*datasets[:2], names=["DS1", "DS2"])

# Change the initial search window
# Automatically applies the opposite to DS2->DS1
a.set_coarse_params({"DS1": {"DS2": {"RT": {"upper": 0.5, "lower": -0.01}}}}, rec=True)

# Automatically generates all anchors for every dataset
a.gen_anchors()

# Generate coarse matches
a.gen_coarse()

# Generate the scalers. Will not overwrite custom scalers
a.gen_scalers()

# Find standard deviations
a.gen_stds()

# Scale the feature descriptors
a.gen_scaled_values()

# Generate subalignment matches
a.gen_matches()

# generates a.graph, a network-x graph
a.gen_graph()


############
# Example 3 - One way supervised alignment to another dataset
############
a = MSAligner(*datasets, names=names)

# Automatically generates all anchors for every dataset
a.gen_anchors()

# Creates a coarse result of anchors from DS1 to DS2
a.gen_coarse("DS1", "DS2")

# Create the scalers for DS1 to DS2
a.gen_scalers("DS1", "DS2")

# Calculate the standard deviations
a.gen_stds("DS1", "DS2")

# Scale the feature descriptors
a.gen_scaled_values("DS1", "DS2")

# Scales DS1 to DS2 based on generated scalers
a.gen_scaled_values("DS1", "DS2")

a.gen_matches("DS1", "DS2")

a.report(
    datasets_1="DS1",
    datasets_2="DS2",
    filepath="one-way-results.pdf",
)


############
# Example 4 - Two way unsupervised alignment to one or more other datasets
############
a = MSAligner(*datasets, names=["DS1", "DS2", "DS3"])

# Align the
a.align("DS1", ["DS2", "DS3"])
a.align(["DS2", "DS3"], "DS1")

a.report(datasets_1="DS1", datasets_2=["DS2", "DS3"], filepath="DS1_to.pdf")
a.report(
    datasets_1=["DS2", "DS3"],
    datasets_2="DS1",
    filepath="DS1_from.pdf",
)

a.to_csv("DS1", g_size_or_loss=1, c_size_or_loss=1, remove_rerank=False)
